
#include "FileReader.h"

// constructor
FileReader::FileReader ()
{
	bFileOpened = false;
	characterSet = FileReader::ASCII;
}

// destructor
FileReader::~FileReader ()
{
	Close ();
}

// tests whethr a file with given name exists or not
bool FileReader::DoesExist(const char* cFileNamePtr)
{
	ifstream iTestFile;
	iTestFile.open (cFileNamePtr);
	bool bRetStatus = ((iTestFile) ? true : false);
	iTestFile.close ();
	return bRetStatus;
}

// opens given ascii or unicode file for input
bool FileReader::Open (const char* cFileNamePtr, CODE characterSet)
{
	// check if invaid file name
	if (cFileNamePtr == NULL || strcmp(cFileNamePtr, "") == 0)
		throw IOException (__FILE__, __LINE__, FILE_NOT_FOUND, 
			"Invalid file name specified.");

	// close previously opened file
	if (bFileOpened)
		Close ();

	ifsReader.open (cFileNamePtr, ios::binary);
	if (!ifsReader)
	{
		Close();
		bFileOpened = false;
		string errorMsg = "Could not open file: ";
		errorMsg += cFileNamePtr;
		errorMsg += ".";
		throw IOException (__FILE__, __LINE__, FILE_NOT_FOUND, errorMsg);
	}

	this->characterSet = characterSet;

	// read BOM to check if it is unicode file
	if (characterSet == FileReader::UNI)
	{
		Word ucs_2_bom = 0;
		DWord ucs_4_bom = 0;
		wchar_t ucs_x_bom = 0;

		Byte* ucs_x_bom_ptr = ( (__ULE_UCS_SIZE == __ULE_UCS_2) ? (Byte*)(&ucs_2_bom) : (Byte*)(&ucs_4_bom) );

		ifsReader.read((char*)ucs_x_bom_ptr, __ULE_UCS_SIZE);

		ucs_x_bom = ( (__ULE_UCS_SIZE == __ULE_UCS_2) ? (wchar_t)(*(Word*)ucs_x_bom_ptr) : (wchar_t)(*(DWord*)ucs_x_bom_ptr) );
		if (ucs_x_bom != __ULE_UCS_BOM)
		{
			Close();
			throw IOException (__FILE__, __LINE__, FILE_NOT_FOUND, 
				"File is not Unicode based.");
		}
	}

	bFileOpened = true;

	return true;
}

// reads unicode character string from unicode file
bool FileReader::Read (wstring& wsBuffer, int intNumCh)
{
	// check if file is open
	if (!bFileOpened)
		return false;

	// check if file is end
	if (ifsReader.eof ())
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"End of file encountered.");
		//return false;

	// check if file was opened as unicode file
	if (characterSet != FileReader::UNI)
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"File not openned as Unicode initially.");	// file was not opened as unicode file

	Word wc_2 = 0;
	DWord wc_4 = 0;
	wchar_t wc_x = 0;

	Byte* wc_ptr = ( (__ULE_UCS_SIZE == __ULE_UCS_2) ? (Byte*)(&wc_2) : (Byte*)(&wc_4) );

	int i = 0;

	wstring wsString;

	while (i < intNumCh)
	{
		ifsReader.read ((char*)wc_ptr, __ULE_UCS_SIZE);

		if (ifsReader.eof ())
			break;

		wc_x = ( (__ULE_UCS_SIZE == __ULE_UCS_2) ? (wchar_t)(*(Word*)wc_ptr) : (wchar_t)(*(DWord*)wc_ptr) );
		wsString += wc_x;

		i++;
	}

	// check if could it consumed a single char
	if (wsString == String2Wstring(""))
		return false;

	wsBuffer = wsString;
	return true;
}

// reads a chunk of specified num of bytes from ascii file
bool FileReader::Read(char* buf, int size)
{
	if (buf == NULL || size <= 0)
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"Null buffer or zero size specified.");

	// check if file is open
	if (!bFileOpened)
		return false;

	// check if file is end
	if (ifsReader.eof ())
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"End of file encountered.");
		//return false;

	ifsReader.read ((char*) buf, size);

	// chaeck if could it consumed a single char
	if (ifsReader.gcount() == 0)
		return false;

	return true;
}

// reads a chunk of specified num of CHARs from unicode file (UCS-2 or UCS-4)
bool FileReader::Read(wchar_t* buf, int size)
{
	if (buf == NULL || size <= 0)
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"Null buffer or zero size specified.");

	// check if file is open
	if (!bFileOpened)
		return false;

	// check if file is end
	if (ifsReader.eof ())
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"End of file encountered.");
		//return false;

	Word wc_2 = 0;
	DWord wc_4 = 0;
	wchar_t wc_x = 0;

	Byte* wc_ptr = ( (__ULE_UCS_SIZE == __ULE_UCS_2) ? (Byte*)(&wc_2) : (Byte*)(&wc_4) );

	// reset buf
	buf[0] = 0;

	int i = 0;
	while (i < size)
	{
		ifsReader.read ((char*)wc_ptr, __ULE_UCS_SIZE);

		if (ifsReader.eof ())
			break;

		wc_x = ( (__ULE_UCS_SIZE == __ULE_UCS_2) ? (wchar_t)(*(Word*)wc_ptr) : (wchar_t)(*(DWord*)wc_ptr) );
		buf[i] = wc_x;
		i++;
	}

	// check if it could consume at least 1 char
	if (i == 0)
		return false;

	return true;
}

// reads a chunk of specified num of bytes from ascii file from specifed location
bool FileReader::Read(char* buf, int from, int size)
{
	if (buf == NULL || size <= 0)
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"Null buffer or zero size specified.");

	// check if file is open
	if (!bFileOpened)
		return false;

	// check if file is end
	if (ifsReader.eof ())
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"End of file encountered.");
		//return false;

	ifsReader.clear();	// reset reader object if it pass end of file
	SeekG(from);

	ifsReader.read ((char*) buf, size);

	// check if could it consumed a single char
	if (ifsReader.gcount() == 0)
		return false;

	return true;
}

// reads ascii character string from ascii file until delim or eof found
bool FileReader::GetLine (string& sBuffer, int intNumCh, char cDelim)
{
	// check if file is open
	if (!bFileOpened)
		return false;

	// check if file is end
	if (ifsReader.eof ())
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"End of file encountered.");
		//return false;

	// check if file was opened as unicode file
	if (characterSet != FileReader::ASCII)
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"File not openned as ASCII initially.");	// file was not opened as unicode file

	char c;
	int i = 0;

	string sString;

	while (i < intNumCh)
	{
		ifsReader.read ((char*)&c, 1);
		
		if (ifsReader.eof ())
			break;

		if (c == cDelim) // delim found
			break;

		sString += c;

		i++;
	}

	// check if could it consumed a single char
	if (sString == "")
		return false;
	
	sBuffer = sString;
	
	return true;
}

// reads unicode character string from unicode file until delim or eof found
bool FileReader::GetLine (wstring& wsBuffer, int intNumCh, wchar_t wcDelim)
{
	// check if file is open
	if (!bFileOpened)
		return false;

	// check if file is end
	if (ifsReader.eof ())
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"End of file encountered.");
		//return false;

	// check if file was opened as unicode file
	if (characterSet != FileReader::UNI)
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"File not openned as UNICODE initially.");	// file was not opened as unicode file

	Word wc_2 = 0;
	DWord wc_4 = 0;
	wchar_t wc_x = 0;

	Byte* wc_ptr = ( (__ULE_UCS_SIZE == __ULE_UCS_2) ? (Byte*)(&wc_2) : (Byte*)(&wc_4) );

	int i = 0;

	wstring wsString;

	while (i < intNumCh)
	{
		ifsReader.read ((char*)wc_ptr, __ULE_UCS_SIZE);
		
		if (ifsReader.eof ())
			break;

		wc_x = ( (__ULE_UCS_SIZE == __ULE_UCS_2) ? (wchar_t)(*(Word*)wc_ptr) : (wchar_t)(*(DWord*)wc_ptr) );

		if (wc_x == wcDelim) // delim found
			break;

		wsString += wc_x;

		i++;
	}

	// check if could it consumed a single char
	if (wsString.empty())
		return false;
	
	wsBuffer = wsString;
	
	return true;
}

// puts file pointer to specified number of ascii chars back
bool FileReader::UngetAsciiCharacters(int intNumCh)
{
	// check if file is open
	if (!bFileOpened)
		return false;

	// check if file was opened as ascii file
	if (characterSet != FileReader::ASCII)
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"File not openned as ASCII initially."); // file was not opened as ascii file

	ifsReader.clear();	// reset reader object if it pass end of file
	ifsReader.seekg(-intNumCh, ios::cur);
	return true;
}

// puts file pointer to specified number of unicode chars back
// (2 bytes for UCS-2 and 4 bytes for UCS-4)
bool FileReader::UngetUnicodeCharacters(int intNumCh)
{
	// check if file is open
	if (!bFileOpened)
		return false;

	// check if file was opened as unicode file
	if (characterSet != FileReader::UNI)
		throw IOException (__FILE__, __LINE__, FILE_READ_ERROR, 
			"File not openned as Unicode initially.");	// file was not opened as unicode file

	ifsReader.clear();	// reset reader object if it pass end of file
	ifsReader.seekg(-intNumCh*__ULE_UCS_SIZE, ios::cur);
	return true;
}

bool FileReader::SeekG (int loc)
{
	// check if file is open
	if (!bFileOpened)
		return false;

	if (characterSet == FileReader::UNI)
		ifsReader.seekg(loc+__ULE_UCS_SIZE);
	else if (characterSet == FileReader::ASCII)
		ifsReader.seekg(loc);

	return true;
}

int FileReader::FileSize ()
{
	// check if file is open
	if (!bFileOpened)
		return false;

	int cur_loc = ifsReader.tellg();
	ifsReader.seekg (0, ios::end);
	int last_byte = ifsReader.tellg();
	ifsReader.seekg (cur_loc, ios::beg);

	return last_byte;
}

// closes file if opened
void FileReader::Close ()
{
	if (bFileOpened)
	{
		ifsReader.close ();
		ifsReader.clear ();
		bFileOpened = false;
	}
}

// checks whether file has end
bool FileReader::EoF ()
{
	// check if file is open
	if (!bFileOpened)
		return false;

	return ifsReader.eof ();
}
