#include <iostream.h>

#include "HashTable.h"



HashTable::HashTable()
{
	arrSize = tableSize = filledElementCount = 0;
	lastEmptyNode = -1; // to include zeroth index also
	hTable = NULL;
	hf = NULL;
	cf = NULL;

	return;
}


bool HashTable::Allocate(int arrSize, int tableSize)
{
	// de-allocate previous hash table (if any)
	Deallocate();

	// check validity to given sizes
	if (arrSize < 1 || tableSize < 1)
		return false;
	
	// set new sizes
	this->arrSize = arrSize;
	this->tableSize = ((tableSize == 0 || tableSize > arrSize) ? arrSize : tableSize);

	// allocate new space for new hash table
	hTable = new HashNode [arrSize];
	if (!hTable)
		return false;

	// initialize elements of new hash table
	for (int i = 0; i < arrSize; i++)
		hTable[i].Next = HTEMPTY;

	return true;
}


void HashTable::Deallocate()
{
	if (hTable != NULL)
	{
		delete [] hTable;
		hTable = NULL;
	}

	arrSize = tableSize = filledElementCount = 0;
	lastEmptyNode = -1;
	hf = NULL;
	cf = NULL;

	return;
}


HashTable::~HashTable()
{
	Deallocate ();
}


bool HashTable::Build(string wordlistFilePath, UniChar2IndexMap &charmap)
{

	FileReader wordlistReader;
	
	wstring tempwordcount,tempword,wfreq; 
	unsigned long wordcount=0,wordsAdded=0; 
	int freq;
	

	try{
		wordlistReader.Open(wordlistFilePath.c_str(),FileReader::UNI);
	}
	catch (IOException& ex)
	{	// set params telling "where" actaully it caused exception
		ex.SetFileName (__FILE__);
		ex.SetLineNum (__LINE__);
		ex.SetExceptionMsg (ex.GetExceptionMsg() + " Config file not opened sucessfully.");
		throw; // re-throw the same exception object
	}
	try{
		wordlistReader.GetLine(tempwordcount,7,'\n');
		wordcount = wstring2int(tempwordcount)-1;
		if(wordcount<=1) 
			return false;

		if (Allocate(wordcount,wordcount)==false)
			return false;

		wordlistReader.GetLine(tempword,22,'\t');
		wordlistReader.GetLine(wfreq,8,'\n');
		freq = wstring2int(wfreq);
		
	//	Trim(tempword);
		tempword = charmap.Validate(tempword);
		while(!wordlistReader.EoF() && wordsAdded<wordcount)
		{

			if(Insert(tempword,freq)==HTEMPTY)
				Prob(tempword,freq,Hash(tempword));

			wordsAdded++;

			wordlistReader.GetLine(tempword,22,'\t');
			wordlistReader.GetLine(wfreq,8,'\n');
			freq = wstring2int(wfreq);
	//		Trim(tempword);
			tempword = charmap.Validate(tempword);

			if(wordsAdded==149464)
				cout<<"\n"<<wordsAdded<<flush;
		}
			
		wordlistReader.Close();
	}
	catch (IOException& ex)
	{	// set params telling "where" actaully it caused exception
		ex.SetFileName (__FILE__);
		ex.SetLineNum (__LINE__);
		ex.SetExceptionMsg (ex.GetExceptionMsg() + " Config file not opened sucessfully.");
		throw; // re-throw the same exception object
	}


	return true;
}



//##ModelId=4151595A0510
bool HashTable::Save(string filename)
{
	FileWriter hashWriter;
	DWord i;
	wstring tword;
	HashNode hnode;
	 

	try{
		hashWriter.Open(filename.c_str(),FileWriter::UNI);
	}
	catch (IOException& ex)
	{	// set params telling "where" actaully it caused exception
		ex.SetFileName (__FILE__);
		ex.SetLineNum (__LINE__);
		ex.SetExceptionMsg (ex.GetExceptionMsg() + " Config file not opened sucessfully.");
		throw; // re-throw the same exception object
	}

	try{
		
		hashWriter.Write((char*)&filledElementCount,sizeof(DWord));

		for(i=0; i<filledElementCount; i++)
		{
			this->GetNodeAtIndex(i,hnode);
			
			tword.resize(20,L'\0');
			for (int j = 0; j < 20; j++)
			{
				if (j < hnode.HashKey.size())
					tword[j] = hnode.HashKey[j];
					else
					tword[j] = 0;
			}

			hashWriter.Write(tword);
			hashWriter.Write((char*)&hnode.Frequency,sizeof(int));
			hashWriter.Write((char*)&hnode.Next,sizeof(DWord));

		}
		hashWriter.Close();
	}
	catch (IOException& ex)
	{	// set params telling "where" actaully it caused exception
		ex.SetFileName (__FILE__);
		ex.SetLineNum (__LINE__);
		ex.SetExceptionMsg (ex.GetExceptionMsg() + " Config file not opened sucessfully.");
		throw; // re-throw the same exception object
	}


	return true;
}

//##ModelId=4151595A050E
bool HashTable::Load(string filename)
{

	FileReader hashReader;
	DWord hashArrayCount,i,k;
	HashNode hnode;

	wstring tword;


//	this->Deallocate();

	try{
		hashReader.Open(filename.c_str(),FileReader::UNI);
	}
	catch (IOException& ex)
	{	// set params telling "where" actaully it caused exception
		ex.SetFileName (__FILE__);
		ex.SetLineNum (__LINE__);
		ex.SetExceptionMsg (ex.GetExceptionMsg() + " Config file not opened sucessfully.");
		throw; // re-throw the same exception object
	}

		
	try{
	
		hashReader.Read((char*)&hashArrayCount,sizeof(DWord));

		if (Allocate(hashArrayCount,hashArrayCount) == false)
			return false; // could not allocate space for hash table

		for(i=0;i<hashArrayCount && !hashReader.EoF() ;i++)
		{
			hashReader.Read(tword,20);
			hashReader.Read((char*)&hnode.Frequency,sizeof(int));
			hashReader.Read((char*)&hnode.Next,sizeof(DWord));

			//truncate null characters at the end of read string.
			for(k=0;k<20;k++)
			{
				if(tword[k]==0x0000)
					break;	
			}

			hnode.HashKey= tword.substr(0,k);			

			this->SetNodeAtIndex(i,hnode);

//			GetNodeAtIndex(i,hnode);
		}

		hashReader.Close();
	}
	catch (IOException& ex)
	{	// set params telling "where" actaully it caused exception
		ex.SetFileName (__FILE__);
		ex.SetLineNum (__LINE__);
		ex.SetExceptionMsg (ex.GetExceptionMsg() + " Config file not opened sucessfully.");
		throw; // re-throw the same exception object
	}


	return true;
	
}


//##ModelId=4151595A0517
int HashTable::Insert(wstring key,int freq)
{
	if (hTable == NULL)
		// not initialized
		return -1;

	unsigned long h = 0;

	h = Hash(key);

	if (hTable[h].Next == HTEMPTY)
	{
		hTable[h].HashKey = key;
		hTable[h].Frequency = freq;		
		hTable[h].Next = HTNIL;
		filledElementCount++;
	}
	else
	{
		return -1;
	}

	return h;
}


//##ModelId=4151595A0519
bool HashTable::Search(wstring str, vector <HashNode> & res)
{
	// clear out parameter
	res.clear();


	if (hTable == NULL)
		// not initialized
		return false;

	int h = Hash(str);

	if(hTable[h].Next == HTEMPTY)
		return false;

	HashNode hn;

	while (h != HTNIL)
	{
		hn = hTable[h];

		// compare strings
		if ((cf != NULL && cf(hn.HashKey, str) == true) 
			|| (cf == NULL && hn.HashKey == str))
		{
			if (hn.HashKey.length() > 0 && 
					hn.HashKey[0] != __HT_DELETED_NODE) // skip entries marked deleted
			{
				hn.Index = h;
				res.push_back(hn);
			}
		}

		// try next chain element
		h = hTable[h].Next;
		if(h == HTEMPTY)
			return false;
	}


	return ( (res.size() > 0) ? true : false );
}

bool HashTable::Search(wstring str)
{
	
	if (hTable == NULL)
		// not initialized
		return false;

	int h = Hash(str);

	if(hTable[h].Next == HTEMPTY)
		return false;

	HashNode hn;

	while (h != HTNIL)
	{
		hn = hTable[h];

		// compare strings
		if ((cf != NULL && cf(hn.HashKey, str) == true) 
			|| (cf == NULL && hn.HashKey == str))
		{
			if (hn.HashKey.length() > 0 && 
					hn.HashKey[0] != __HT_DELETED_NODE) // skip entries marked deleted
			{
				return true;
			}
		}

		// try next chain element
		h = hTable[h].Next;
		if(h == HTEMPTY)
			return false;
	}


	return false;
}

//##ModelId=4151595A051C
bool HashTable::GetNodeAtIndex(int index, HashNode& hn)
{
	if (hTable == NULL)
		// not initialized
		return false;

	if (index < 0 || index >= arrSize)
		// out of bound entry accessed
		return false;

	hn = hTable[index];
	return true;
}


bool HashTable::SetNodeAtIndex(int index, HashNode& hn)
{
	if (hTable == NULL)
		// not initialized
		return false;

	if (index < 0 || index >= arrSize)
		// out of bound entry accessed
		return false;

	hTable[index].HashKey = hn.HashKey;
	hTable[index].Frequency = hn.Frequency;
	hTable[index].Next = hn.Next;

	return true;
}


bool HashTable::DeleteNode(int index)
{
	if (hTable == NULL)
		// not initialized
		return false;

	if (index < 0 || index >= arrSize)
		// out of bound entry accessed
		return false;

	if (hTable[index].Next != HTEMPTY)
	{
		int i = 0;
		if (hTable[index].HashKey.length() == 0)
			hTable[index].HashKey.insert (i, 1, __HT_DELETED_NODE);
		else
			hTable[index].HashKey[i] = __HT_DELETED_NODE;
	}

	return true;
}


//##ModelId=4151595A0512
DWord HashTable::Hash(wstring key)
{
	unsigned long h = 0, a = 31415, b = 27183;
	if(arrSize > 0)
	{
		if (hf != NULL) // call specialized func if specified
			h = hf(key, tableSize);
		else
		{
			for(int i = 0; i < key.size(); i++, a = a*b%(tableSize - 1))
			{
				wchar_t wc = key[i];
				h = (a*h + key[i]) % (tableSize);
			}
		}
	}
	
	return h;
}


//##ModelId=4151595A0514
/////////////////// temp code
int HashTable::Prob(wstring key,int freq, DWord hashCode)
{
	if (hTable == NULL)
		// not initialized
		return -1;

	if (hashCode < 0 || hashCode >= tableSize)
		return -1;

	unsigned long index = hashCode;
	unsigned long last;
	while (hTable[index].Next != HTNIL)
	{
		index = hTable[index].Next;
	}

	last = index;
	//index = hashSize;
	index = lastEmptyNode; 
	//index = 0; // start searching from 0th index
	do {
		index++;
		if (index >= arrSize)	// no more space to add
			return -1;
	} while (hTable[index].Next != HTEMPTY);

	hTable[last].Next = index;
	hTable[index].HashKey = key;
	hTable[index].Frequency = freq;
	hTable[index].Next = HTNIL;
	filledElementCount++;

	lastEmptyNode = index;

	return index;
}


int HashTable::getArrSize(){
	return arrSize;
}


DWord HashTable::elementCount(){
	return filledElementCount;
}


void HashTable::SetSepcialFunctions(HashFunc h, CompareFunc c)
{
	hf = h;
	cf = c;
}

